"""
Module used to construct SQL queries that will be used to query the database.

Instead of each individual module building the queries based on their provided inputs,
use this module to manage this task in a central location.

When each of the primary modules are called, they will call a function here, and using
the provided parameters, an SQL query string will be constructed and returned.
"""

from pyhockey.util.input_validation import check_input_type, check_input_values, validate_date_range

# Define custom type for inputs into our queries
type QueryValue = str | int | float | list[str] | list[int] | list[float]


def construct_query(table_name: str,
                    column_mapping: dict[str, QueryValue],
                    qualifiers: dict[str, str] | None = None,
                    order_by: list[str] | None = None) -> str:
    """
    Method that takes parameters passed into the primary functions and constructs an
    SQL query that can be used to query the data.

    :param str table_name: The name of the table being queried.
    :param dict[str, QueryValue] column_mapping: A dict mapping column names in the table to values
                                                 they need to be evaluated against. Multiple values
                                                 can be provided in a list and all will be combined
                                                 in an 'OR' statement.
    :param dict[str, str] qualifiers: A dict mapping certain column names to evaluations which will
                                      be applied to the query, e.g. '<' or '>' conditions, defaults
                                      to None.
    :param list[str] order_by: A list of strings corresponding to column names that the results
                               will be sorted by.

    :return str: The full query provided as a string.
    """

    # This condition will raise an error if mis-matched types or invalid values were provided
    if check_input_type(column_mapping=column_mapping) and \
       check_input_values(column_mapping=column_mapping):
        pass

    query: str = f"SELECT * FROM {table_name} WHERE "

    query_conditions: list[str] = []

    if qualifiers:
        # Qualifiers will be provided in a dict in a format, e.g.,
        #   'iceTime': '>100',
        # to indicate that the query should filter for entries with iceTime > 100.

        # When dealing with date ranges, some additional checks are necessary
        if 'start_date' in qualifiers.keys() or 'end_date' in qualifiers.keys():
            column_mapping = validate_date_range(column_mapping, qualifiers)

        for column_name, value in qualifiers.items():
            # 'start_date' and 'end_date' need slightly special handling, as they don't
            # correspond to actual table columns
            if column_name == 'start_date':
                query_conditions.append(f"gameDate >= '{value}'")
            elif column_name == 'end_date':
                query_conditions.append(f"gameDate <= '{value}'")
            else:
                query_conditions.append(f"{column_name} {value}")

    for column_name, value in column_mapping.items():
        # The keys of the column_mapping dict will be strings corresponding to the column
        # names in the table, whereas the values will be filters applied to those columns.

        # Default value for most columns is None, so skip those
        if value is None:
            continue

        # Also skip the 'team' input if the value is 'ALL', since no filter is needed
        if column_name == 'team' and value == 'ALL':
            continue

        # Names are handled slightly differently than other columns
        if column_name == 'name':
            condition: str = handle_names(value)

        # These values can be of multiple types, and can also potentially be a list of items.
        elif isinstance(value, list):
            if isinstance(value[0], str):
                # Add single quotes to value if dealing with strings
                condition: str = " OR ".join(f"{column_name} = '{v}'" for v in value)
            else:
                condition: str = " OR ".join(f"{column_name} = {v}" for v in value)

            # Make sure the 'OR' conditions are bracketed
            condition = f"({condition})"

        # If we're dealing with a singleton and not a list...
        else:
            # Add single quotes to value if dealing with strings
            if isinstance(value, str):
                condition: str = f"{column_name} = '{value}'"
            else:
                condition: str = f"{column_name} = {value}"

        query_conditions.append(condition)

    all_conditions: str = " AND ".join(query_conditions)

    query += all_conditions

    if order_by:
        # If order_by is provided, it will be a list of column names, so construct
        # the ORDER BY statement and append it to the query.
        order: str = f" ORDER BY {', '.join(order_by)}"

        query += order

    return query


def handle_names(value: str | list[str]) -> str:
    """
    When names are provided to filter a table on, use a 'LIKE' comparison instead of '=' as well
    as wildcards to try and cover as many provided inputs as possible. 
    
    E.g., if just a single word is given, assume it to be either a first or last name, and thus
    the condition will be
        name LIKE '%INPUT%', 
    whereas if two or more words are given, the condition will look like
        name LIKE '%INPUT_A%INPUT_B%'.

    :param str value: The provided input(s) for the name to be filtered against.
    :return str: _description_
    """

    # These values can be of multiple types, and can also potentially be a list of items.
    if isinstance(value, list):
        name_values = [f'%{"%".join(v.split())}%' for v in value]
        condition: str = " OR ".join(f"name LIKE '{n}'" for n in name_values)

        # Make sure the 'OR' conditions are bracketed
        condition = f"({condition})"

    # If we're dealing with a singleton and not a list...
    else:
        name_value = f'%{"%".join(value.split())}%'
        condition: str = f"name LIKE '{name_value}'"

    return condition



import pytest

from pyhockey.util.query_builder import construct_query, handle_names


def test_construct_query_simple():
    """
    Test the construction of a simple query with one string filter and one int filter.
    """
    conditions: dict[str] = {
        'team': 'TOR',
        'season': 2025,
        'situation': '5on5'
    }

    result: str = construct_query(table_name='skaters', column_mapping=conditions)

    expected: str = "SELECT * FROM skaters WHERE team = 'TOR' AND season = 2025 AND "\
                    "situation = '5on5'"

    assert result == expected


def test_construct_query_complex():
    """
    Test the construct of a more complex query with multiple filter values on columns, as well
    as a qualifier.
    """
    conditions: dict[str] = {
        'team': ['TOR', 'MTL', 'OTT'],
        'season': [2024, 2025],
        'situation': ['all', '5on5'],
        'name': ['Stolarz', 'Sam Montembault']
    }

    qualifiers: dict[str] = {
        'iceTime': '<100'
    }

    result: str = construct_query(table_name='goalies', column_mapping=conditions,
                                  qualifiers=qualifiers, order_by=['team', 'season'])

    expected: str = "SELECT * FROM goalies WHERE "\
                    "iceTime <100 "\
                    "AND (team = 'TOR' OR team = 'MTL' OR team = 'OTT') "\
                    "AND (season = 2024 OR season = 2025) "\
                    "AND (situation = 'all' OR situation = '5on5') "\
                    "AND (name LIKE '%Stolarz%' OR name LIKE '%Sam%Montembault%') "\
                    "ORDER BY team, season"

    assert result == expected


def test_construct_query_fails_with_bad_input_type():
    """
    That that construct_query() raises a ValueError if given a column mapping with mis-matched
    input types.
    """
    conditions: dict[str] = {
        'team': 'TOR',
        'season': '2025'
    }

    with pytest.raises(ValueError):
        construct_query(table_name='skaters', column_mapping=conditions)


def test_construct_query_with_date_range():
    """
    Test that construct_query() works as expected when providing date ranges.
    """
    conditions: dict[str] = {
        'team': 'TOR',
        'season': [2024],
        'situation': ['all', '5on5'],
    }

    qualifiers: dict[str] = {
        'start_date': '2024-10-30',
        'end_date': '2025-03-25'
    }

    result: str = construct_query(table_name='team_games', column_mapping=conditions,
                                  qualifiers=qualifiers, order_by=['team', 'gameDate'])

    expected: str = "SELECT * FROM team_games WHERE "\
                    "gameDate >= '2024-10-30' "\
                    "AND gameDate <= '2025-03-25' "\
                    "AND team = 'TOR' "\
                    "AND (situation = 'all' OR situation = '5on5') "\
                    "ORDER BY team, gameDate"

    assert result == expected


def test_handle_names_singleton():
    """
    Test that handle_names works as expected when a single name is provided, either a full
    name or just a last name.
    """
    condition: str = handle_names('Matthews')
    assert condition == "name LIKE '%Matthews%'"

    condition_b: str = handle_names('Auston Matthews')
    assert condition_b == "name LIKE '%Auston%Matthews%'"


def test_handle_names_list():
    """
    Test that handle_names works as expected when a list of names is provided.
    """
    condition: str = handle_names(['Matthews', 'William Nylander'])
    assert condition == "(name LIKE '%Matthews%' OR name LIKE '%William%Nylander%')"



from .goalie_summary import goalie_summary
from .skater_summary import skater_summary
from .team_summary import team_summary
from .team_games import team_games
from .skater_games import skater_games
from .goalie_games import goalie_games



"""
Main module for returning game-by-game statistics for each skater.
"""
import polars as pl

from pyhockey.util.query_table import query_table
from pyhockey.util.data_disclaimer import print_data_disclaimer


# Define custom type for inputs into our queries
type QueryValue = str | int | float | list[str] | list[int] | list[float]


def skater_games(season: int | list[int] | None = None,
                 name: str | list[str] | None = None,
                 team: str | list[str] = 'ALL',
                 start_date: str | None = None,
                 end_date: str | None = None,
                 situation: str = 'all',
                 quiet: bool = False) -> pl.DataFrame:
    """
    Primary function for returning game-by-game skater statistics. Accepts one or multiple teams, 
    one or multiple skater names, one or multiple seasons, or alternatively, a start- and end-date
    for which to return game-by-game metrics.

    :param int | list[int] | None season: Either one or a list of seasons for which to return all
                                          games. Disregarded if both start_date and end_date are
                                          also provided, defaults to None
    :param str | list[str] | None name: Either one or a list of names for which to return stats. 
                                        Can be a full name, partial name, or just first/last name.
                                        Defaults to None.
    :param str | list[str] team: Either one or a list of teams, provided in 3-letter acronyms, 
                                 defaults to 'ALL'
    :param str | None start_date: A date from which to return all games on and after that date, in
                                  YYYY-MM-DD format, defaults to None
    :param str | None end_date: A date from which to return all games before and on that date, in
                                YYYY-MM-DD format, defaults to None
    :param str situation: One of 'all', '5on5', '4on5', or '5on4', defaults to 'all'
    :param bool quiet: If set to True, don't print the data disclaimer, defaults to False
    
    :raises ValueError: Raises a ValueError if incorrect date format is provided.
    
    :return pl.DataFrame: A DataFrame containing the requested data.
    """

    if not season and not start_date and not end_date:
        raise ValueError("No values provided for 'season', 'start_date', or 'end_date'. Must "\
                         "provide value for at least one of these.")

    qualifers: dict[str, str] = {}

    if start_date:
        qualifers['start_date'] = start_date
    if end_date:
        qualifers['end_date'] = end_date

    column_mapping: dict[str, QueryValue] = {
        'season': season,
        'name': name,
        'team': team,
        'situation': situation
    }

    results: pl.DataFrame = query_table(table='skater_games', column_mapping=column_mapping,
                                        qualifiers=qualifers, order_by=['team', 'gameDate'])

    if not quiet:
        print_data_disclaimer(source='NaturalStatTrick')

    return results



"""
Module for checking that provided inputs are of the correct type as expected by the database 
schema, and also checking that the values provided are valid in the context of what the
database expects (i.e. using the proper team acronyms and seasons for which data is
available).
"""

import datetime
from datetime import datetime


# Define custom type for inputs into our queries
type QueryValue = str | int | float | list[str] | list[int] | list[float]


### CONSTANTS #####################################################################################

# The valid values accepted for each column
VALID_SITUATIONS = ['all', '5on5', '4on5', '5on4', 'other']

VALID_SEASONS = list(range(2008, 2026))

VALID_TEAMS = ['ANA', 'ARI', 'ATL', 'BOS', 'BUF', 'CAR', 'CBJ', 'CGY', 'CHI', 'COL', 'DAL', 'DET',
               'EDM', 'FLA', 'L.A', 'LAK', 'MIN', 'MTL', 'N.J', 'NJD', 'NSH', 'NYI', 'NYR', 'OTT',
               'PHI', 'PIT', 'S.J', 'SEA', 'SJS', 'STL', 'T.B', 'TBL', 'TOR', 'UTA', 'VAN', 'VGK',
               'WPG', 'WSH', 'ALL']

VALID_INPUT_VALUES = {
    'season': VALID_SEASONS,
    'team': VALID_TEAMS,
    'situation': VALID_SITUATIONS
}

# A mapping of column names in the database to the expected types of the values in those columns
COLUMN_SCHEMA = {
    'name': str,
    'gameID': int,
    'gameDate': str,
    'season': int,
    'team': str,
    'state': str,
    'situation': str,
    'iceTime': (int, float),
    'shotsAgainst': int,
    'goalsAgainst': int,
    'xGoalsAgainst': (int, float),
    'gamesPlayed': int,
    'xGoals': (int, float),
    'goals': int,
    'lowDangerShots': int,
    'mediumDangerShots': int,
    'highDangerShots': int,
    'lowDangerxGoals': (int, float),
    'mediumDangerxGoals': (int, float),
    'highDangerxGoals': (int, float),
    'lowDangerGoals': int,
    'mediumDangerGoals': int,
    'highDangerGoals': int,
    'position': str,
    'primaryAssists': int,
    'secondaryAssists': int,
    'shots': int,
    'individualxGoals': (int, float),
    'goalsFor': int,
    'xGoalsFor': (int, float),
    'xGoalsShare': (int, float),
    'corsiFor': int,
    'corsiAgainst': int,
    'corsiShare': (int, float),
    'xGoalsForPerHour': (int, float),
    'xGoalsAgainstPerHour': (int, float),
    'goalsForPerHour': (int, float),
    'goalsAgainstPerHour': (int, float),
    'pointsPerHour': (int, float),
    'goalsPerHour': (int, float),
    'averageIceTime': (int, float),
    'corsiPercentage': (int, float),
}
### END CONSTANTS #################################################################################


def check_input_values(column_mapping: dict[str]) -> bool:
    """
    Function to determine if the input values provided by a user are valid in the context of the
    tables being queried, e.g. make sure whatever season value is provided fits within the range
    of seasons which have available data.

    If inputs are found to contain any invalid values, raise a ValueError with details on what the
    valid inputs are.

    :param dict[str] column_mapping: A mapping of the columns to check with their provided inputs

    :raises ValueError: This function will raise a ValueError, ending the program, if an invalid
                        value is provided.

    :return bool: Returns True if no ValueError is raised
    """

    for column, input_value in column_mapping.items():

        # Default value for most columns is None, so skip those
        if input_value is None:
            continue

        if not VALID_INPUT_VALUES.get(column, False):
            # Some columns don't need to have their inputs validated (e.g. names)
            continue

        valid_inputs: list[str | int] = VALID_INPUT_VALUES[column]

        # If input_value is a singleton check that it's in the list of valid inputs, but
        # if input_value is a list check that it is a subset
        if (not isinstance(input_value, list) and input_value not in valid_inputs) or \
           (isinstance(input_value, list) and not set(input_value).issubset(valid_inputs)):

            # If valid_inputs does not contain input_value, construct the error message and
            # raise the error
            msg: str = f"Invalid input '{input_value}' provided for {column}.\n"\
                       f"Valid inputs are {valid_inputs}"

            raise ValueError(msg)

    return True


def check_input_type(column_mapping: dict[str]) -> bool:
    """
    Validates the types provided to the primary functions to make sure they align with
    database expectations when building the query.

    :param dict[str] column_mapping: A mapping of the columns to check with their provided inputs

    :raises ValueError: This function will raise a ValueError, ending the program, if a mismatched
                        type for the value is provided.

    :return bool: If no error is raised, the function will return True.
    """

    for column_name, value in column_mapping.items():

        # Default value for most columns is None, so skip those
        if value is None:
            continue

        desired_type = COLUMN_SCHEMA[column_name]
        # First make sure values that were supplied are the correct types.
        if not isinstance(value, desired_type):
            if isinstance(value, list):
                for v in value:
                    if not isinstance(v, desired_type):
                        raise ValueError(f"Values provided for {column_name} must be "\
                                         f"{desired_type}, received {type(v)}: {v}")
            else:
                raise ValueError(f"Values provided for {column_name} must be "\
                                 f"{desired_type}, received {type(value)}: {value}")

    return True


def validate_date_range(column_mapping: dict[str, QueryValue],
                        qualifiers: dict[str, str]) -> dict[str, QueryValue]:
    """
    Function to check any date inputs (e.g. 'start_date', 'end_date') where given in the
    expected format, and to make sure there are no conflicts between these inputs and the
    'season' input.

    :param dict[str, QueryValue] column_mapping: The column mapping provided for the query.
    :param dict[str, str] qualifiers: The qualifiers provided for the query.

    :raises ValueError: Raises a ValueError if a date value was provided in a format that
                        isn't YYYY-MM-DD, or if end_date is not after start_date.

    :return dict[str, QueryValue]: An updated version of the column mapping.
    """

    # First checks that both 'start_date' and 'end_date', if provided, match the YYYY-MM-DD format
    # using datetime.strptime().
    for date in ['start_date', 'end_date']:
        if date in qualifiers.keys():
            date_string: str = qualifiers[date]
            try:
                datetime.strptime(date_string, '%Y-%m-%d')
            except ValueError as e:
                raise ValueError(f"'{date}' provided in unsupported format. Must be YYYY-MM-DD. "\
                                 f"Recieved {date_string}.") from e

    # Check that, if both start_date and end_date were provided, that start_date comes before
    # end_date
    if qualifiers.get('end_date', False) and qualifiers.get('start_date', False):
        start = datetime.strptime(qualifiers['start_date'], '%Y-%m-%d')
        end = datetime.strptime(qualifiers['end_date'], '%Y-%m-%d')
        if end < start:
            raise ValueError(f"Provided end date ({qualifiers['end_date']}) comes before "\
                             f"provided start date ({qualifiers['start_date']}). Start date "\
                              "must be before or equal to end date.")

    # Then, if both 'start_date' and 'end_date' were provided in addition to 'season', remove
    # 'season' from the column_mapping and print the reason why.
    if column_mapping.get('season', False) and qualifiers.get('start_date', False) and \
        qualifiers.get('end_date', False):
        print("Input values were provided for 'start_date', 'end_date', and 'season'. "\
              "Disregarding the input for 'season' and returning all games between "\
              f"{qualifiers['start_date']} and {qualifiers['end_date']}.")

        del column_mapping['season']

    return column_mapping



import polars as pl

from pyhockey.goalie_summary import goalie_summary


def test_standard_goalie_summary():
    """
    Test that a standard request from goalie_summaries gives a DF of the proper shape.    
    """
    result: pl.DataFrame = goalie_summary(season=[2023, 2024], team=['TOR', 'MTL'],
                                            min_games_played=10)

    assert result.shape == (10, 18)


def test_combined_goalie_summary():
    """
    Test that a request using combined_seasons = True gives a DF of the proper shape.
    """
    result: pl.DataFrame = goalie_summary(season=[2023, 2024], team=['TOR', 'MTL'],
                                          min_games_played=10,
                                          combine_seasons=True)

    assert result.shape == (7, 18)



import polars as pl

from pyhockey.goalie_games import goalie_games


def test_standard_goalie_games():
    """
    Test that a standard request from goalie_games gives a DF of the expected shape.
    """

    result: pl.DataFrame = goalie_games(name=['Stolarz', 'Shesterkin'], start_date='2025-10-01',
                                        end_date='2025-10-25')

    assert result.shape == (14, 10)



"""
Module for creating and managing the connection to the database.

Using duckdb and MotherDuck as our database for now, but no guarantee that this will be the 
case in the future.

A MotherDuck authorization token with read-only access is set in util/config.py, which we use
to connect.
"""

import os
import duckdb
from pyhockey.util.config import MOTHERDUCK_READ_TOKEN


def create_connection(db_name: str = 'md:') -> duckdb.DuckDBPyConnection:
    """
    Creates the connection object to the database.

    :param _type_ db_name: Name of the database to connect to, defaults to 'md:' which uses
                           the authorization token (which is set as an env var) to auto-
                           resolve the database name.

    :return duckdb.DuckDBPyConnection: The connection object which we can use to query the db.
    """

    # MotherDuck uses the MOTHERDUCK_TOKEN env var to authorize read-only access to the db
    os.environ['MOTHERDUCK_TOKEN'] = MOTHERDUCK_READ_TOKEN

    conn: duckdb.DuckDBPyConnection = duckdb.connect(db_name, read_only=True)

    return conn



import polars as pl

from pyhockey.skater_games import skater_games


def test_standard_skater_games():
    """
    Test that a standard request from skater_games gives a DF of the expected shape.
    """

    result: pl.DataFrame = skater_games(name=['Nylander', 'Matthews'], start_date='2025-10-01',
                                        end_date='2025-10-25')

    assert result.shape == (17, 22)



import polars as pl

from pyhockey.skater_summary import skater_summary


def test_standard_skater_summary():
    """
    Test that a standard request from skater_summaries gives a DF of the proper shape.    
    """
    result: pl.DataFrame = skater_summary(season=[2023, 2024], team=['TOR', 'MTL'],
                                            min_icetime=500)

    assert result.shape == (79, 31)


def test_combined_skater_summary():
    """
    Test that a request using combined_seasons = True gives a DF of the proper shape.
    """
    result: pl.DataFrame = skater_summary(season=[2023, 2024], team=['TOR', 'MTL'],
                                            min_icetime=500,
                                            combine_seasons=True)

    assert result.shape == (53, 31)



"""
Module for simple function that displays the source of the data provided by each request.
"""


def print_data_disclaimer(source: str) -> None:
    """
    Prints a simple data disclaimer depending on the source of the data.

    :param str source: Name of the data source, ATM either MoneyPuck or NaturalStatTrick
    """

    source_map = {
        'MoneyPuck': 'https://moneypuck.com',
        'NaturalStatTrick': 'https://naturalstattrick.com'
    }

    print(f"Data for this query provided by {source} ({source_map[source]}).")



"""
Module for handling the actual querying of tables in the database. A lot of the logic for querying
is similar between the primary methods, so made sense to just have it abstracted out and handled
seperately, here.
"""
from typing import Callable

import duckdb
import polars as pl
import polars.selectors as sc

from pyhockey.util.db_connect import create_connection
from pyhockey.util.query_builder import construct_query

# Define custom type for inputs into our queries
type QueryValue = str | int | float | list[str] | list[int] | list[float]


def query_table(table: str,
                column_mapping: dict[str, QueryValue],
                qualifiers: dict[str, str] | None = None,
                order_by: list[str] | None = None,
                combine_seasons: bool = False) -> pl.DataFrame:
    """
    Function to handle specifics of querying a table, given the desired table and parameters.

    This function will handle calling the query constructer, retrieving the database connection,
    querying the database, combining season values if desired, and rounding the values in the
    results before returning them.

    :param str table: Name of the table to be querying.
    :param dict[str, QueryValue] column_mapping: A dict mapping column names in the table to values
                                                 they need to be evaluated against. Multiple values
                                                 can be provided in a list and all will be combined
                                                 in an 'OR' statement.
    :param dict[str, str] qualifiers: A dict mapping certain column names to evaluations which will
                                      be applied to the query, e.g. '<' or '>' conditions, defaults 
                                      to None.
    :param list[str] | None order_by: A list of strings corresponding to column names that the 
                                     results will be sorted by, defaults to None.
    :param bool combine_seasons: , defaults to False
    :return pl.DataFrame: _description_
    """

    query: str = construct_query(table_name=table, column_mapping=column_mapping,
                                 qualifiers=qualifiers, order_by=order_by)

    connection: duckdb.DuckDBPyConnection = create_connection()

    results: pl.DataFrame = connection.sql(query).pl()


    if combine_seasons:
        # A dict mapping table names to functions which can be called to combine season rows into
        # single rows
        season_combine_funcs: dict[str, Callable[[pl.DataFrame], pl.DataFrame]] = {
            'teams': combine_team_seasons,
            'skaters': combine_skater_seasons,
            'goalies': combine_goalie_seasons
        }

        if not isinstance(column_mapping['season'], list):
            print(f"The 'combine_seasons' parameter has been set to 'True', but data for only one "\
                  f"season ({column_mapping['season']}) was requested. Returning data for just "\
                  f"that season...")
            return results
        func: Callable[[pl.DataFrame], pl.DataFrame] = season_combine_funcs[table]
        results = func(results)

    # Round all float values to 2 decimal places before returning
    results = results.with_columns(sc.float().round(2))

    # Close the connection after query is complete
    connection.close()

    return results


def combine_team_seasons(df: pl.DataFrame) -> pl.DataFrame:
    """
    Called when a user requests multiple seasons worth of data and wants to have them combined
    into a single row for each team.

    Goes through the data provided by the query and combines the data for each team-season into
    one row, returning the resulting DataFrame.

    :param pl.DataFrame df: The raw results of the query containing rows for each team-season

    :return pl.DataFrame: The output DataFrame with all team-seasons combined into one row.
    """
    # This list will contain DFs for each individual player, to be concatenated at the end
    team_dfs: list[pl.DataFrame] = []

    # For each unique player, create a filtered DF of just their data and use it to create
    # a dict summarizing the info.
    for team in set(df['team']):
        t_df: pl.DataFrame = df.filter(pl.col('team') == team)
        seasons: list[int] = list(set(t_df['season']))
        seasons.sort()

        # First add the values which are constants.
        combined_info: dict[str] = {
                # The season column will contain each season for this data
                'season': ','.join([str(s) for s in seasons]),
        }

        for col in ['team', 'situation']:
            combined_info[col] = list(t_df[col])[0]

        # Then add the values which are sum totals for each season
        for col in ['gamesPlayed', 'iceTime', 'xGoalsFor', 'goalsFor', 'xGoalsAgainst',
                    'goalsAgainst']:
            combined_info[col] = t_df[col].sum()

        # And finally compute rate metrics from each column containing a total metric value,
        # i.e. goalsFor -> goalsForPerHour (GFph)
        for total_col, rate_col in zip(['goalsFor', 'goalsAgainst', 'xGoalsFor', 'xGoalsAgainst'],
                                       ['goalsForPerHour', 'goalsAgainstPerHour',
                                        'xGoalsForPerHour', 'xGoalsAgainstPerHour']):

            combined_info[rate_col] = combined_info[total_col] * (60.0 / combined_info['iceTime'])

        combined_info['averageIceTime'] = round(combined_info['iceTime'] /
                                                combined_info['gamesPlayed'], 2)

        team_dfs.append(pl.DataFrame(combined_info))

    final_df: pl.DataFrame = pl.concat(team_dfs)

    final_df = final_df.cast(
        {
            'gamesPlayed': pl.Int16,
            'goalsFor': pl.Int16,
            'goalsAgainst': pl.Int16,
            'xGoalsFor': pl.Float32,
            'xGoalsAgainst': pl.Float32,
        }
    )

    return final_df


def combine_skater_seasons(df: pl.DataFrame) -> pl.DataFrame:
    """
    Called when a user requests multiple seasons worth of data and wants to have them combined
    into a single row for each skater.

    Goes through the data provided by the query and combines the data for each player-season into
    one row, returning the resulting DataFrame.

    :param pl.DataFrame df: The raw results of the query containing rows for each player-season

    :return pl.DataFrame: The output DataFrame with all player-seasons combined into one row.
    """
    # This list will contain DFs for each individual player, to be concatenated at the end
    player_dfs: list[pl.DataFrame] = []

    # For each unique player, create a filtered DF of just their data and use it to create
    # a dict summarizing the info.
    for player_id in set(df['playerID']):
        p_df: pl.DataFrame = df.filter(pl.col('playerID') == player_id)
        seasons: list[int] = list(set(p_df['season']))
        seasons.sort()

        if len(seasons) == 1:
            # If the player only has one seasons worth of data in the results, just add that row
            p_df = p_df.cast({'season': pl.String})
            player_dfs.append(p_df)
            continue

        # First add the values which are constants.
        combined_info: dict[str] = {
                'playerID': player_id,
                # The season column will contain each season for this data
                'season': ','.join([str(s) for s in seasons]),
        }

        for col in ['name', 'team', 'position', 'situation']:
            combined_info[col] = list(p_df[col])[0]

        # Then add the values which are sum totals for each season
        for col in ['gamesPlayed', 'iceTime', 'points', 'goals', 'individualxGoals', 'xGoalsFor',
                    'goalsFor', 'xGoalsAgainst', 'goalsAgainst', 'penaltiesTaken',
                    'penaltiesDrawn', 'faceoffsWon', 'faceoffsLost', 'shotsBlocked',
                    'oZoneShifts', 'dZoneShifts', 'neutralZoneShifts', 'flyShifts']:
            combined_info[col] = p_df[col].sum()

        # And finally compute rate metrics from each column containing a total metric value,
        # i.e. goalsFor -> goalsForPerHour (GFph)
        for total_col, rate_col in zip(['goalsFor', 'goalsAgainst', 'xGoalsFor',
                                        'xGoalsAgainst', 'points', 'goals'],
                                        ['goalsForPerHour', 'goalsAgainstPerHour',
                                         'xGoalsForPerHour', 'xGoalsAgainstPerHour',
                                         'pointsPerHour', 'goalsPerHour']):

            combined_info[rate_col] = combined_info[total_col] * (60.0 / combined_info['iceTime'])

        combined_info['averageIceTime'] = round(combined_info['iceTime'] /
                                                combined_info['gamesPlayed'], 2)

        p = pl.DataFrame(combined_info)

        player_dfs.append(p)


    for i, p in enumerate(player_dfs[:]):
        # Make sure dataframe has columns in a consistent order before concat
        p = p[['playerID', 'season', 'name', 'team', 'position', 'situation', 'gamesPlayed',
               'iceTime', 'points', 'goals', 'individualxGoals', 'xGoalsFor', 'xGoalsAgainst',
               'goalsFor', 'goalsAgainst', 'xGoalsForPerHour', 'xGoalsAgainstPerHour',
               'goalsForPerHour', 'goalsAgainstPerHour', 'pointsPerHour', 'goalsPerHour',
               'averageIceTime', 'penaltiesTaken', 'penaltiesDrawn', 'faceoffsWon',
               'faceoffsLost', 'shotsBlocked', 'oZoneShifts', 'dZoneShifts',
               'neutralZoneShifts', 'flyShifts']]

        # Also ensure columns are of consistent types before concat
        p = p.cast(
            {
                'playerID': pl.Int32,
                'gamesPlayed': pl.Int16,
                'iceTime': pl.Float32,
                'points': pl.Int16,
                'goals': pl.Int16,
                'individualxGoals': pl.Float64,
                'goalsFor': pl.Int16,
                'goalsAgainst': pl.Int16,
                'xGoalsFor': pl.Float32,
                'xGoalsAgainst': pl.Float32,
                'pointsPerHour': pl.Float32,
                'goalsPerHour': pl.Float32,
                'goalsForPerHour': pl.Float32,
                'goalsAgainstPerHour': pl.Float32,
                'xGoalsForPerHour': pl.Float32,
                'xGoalsAgainstPerHour': pl.Float32,
                'averageIceTime': pl.Float32,
                'penaltiesTaken': pl.Int16,
                'penaltiesDrawn': pl.Int16,
                'faceoffsWon': pl.Int16,
                'faceoffsLost': pl.Int16,
                'shotsBlocked': pl.Int16,
                'oZoneShifts': pl.Int16,
                'dZoneShifts': pl.Int16,
                'neutralZoneShifts': pl.Int16,
                'flyShifts': pl.Int16
            }
        )

        player_dfs[i] = p
        with pl.Config(tbl_cols=32):
            print(player_dfs[i])

    final_df: pl.DataFrame = pl.concat(player_dfs)

    final_df = final_df.sort(by=['team', 'playerID'])

    return final_df


def combine_goalie_seasons(df: pl.DataFrame) -> pl.DataFrame:
    """
    Called when a user requests multiple seasons worth of data and wants to have them combined
    into a single row for each goalie.

    Goes through the data provided by the query and combines the data for each player-season into
    one row, returning the resulting DataFrame.

    :param pl.DataFrame df: The raw results of the query containing rows for each player-season

    :return pl.DataFrame: The output DataFrame with all player-seasons combined into one row.
    """
    # This list will contain DFs for each individual player, to be concatenated at the end
    player_dfs: list[pl.DataFrame] = []

    # For each unique player, create a filtered DF of just their data and use it to create
    # a dict summarizing the info.
    for player_id in set(df['playerID']):
        p_df: pl.DataFrame = df.filter(pl.col('playerID') == player_id)
        seasons: list[int] = list(set(p_df['season']))
        seasons.sort()

        if len(seasons) == 1:
            # If the player only has one seasons worth of data in the results, just add that row
            p_df = p_df.cast({'season': pl.String})
            player_dfs.append(p_df)
            continue

        # First add the values which are constants.
        combined_info: dict[str] = {
                'playerID': player_id,
                # The season column will contain each season for this data
                'season': ','.join([str(s) for s in seasons]),
        }

        for col in ['name', 'team', 'situation']:
            combined_info[col] = list(p_df[col])[0]

        # Then add the values which are sum totals for each season
        for col in ['gamesPlayed', 'iceTime', 'xGoals', 'goals',
                    'lowDangerShots', 'mediumDangerShots', 'highDangerShots',
                    'lowDangerxGoals', 'mediumDangerxGoals', 'highDangerxGoals',
                    'lowDangerGoals', 'mediumDangerGoals', 'highDangerGoals']:
            combined_info[col] = p_df[col].sum()

        player_dfs.append(pl.DataFrame(combined_info))

    final_df: pl.DataFrame = pl.concat(player_dfs)

    final_df = final_df.cast(
        {
            'gamesPlayed': pl.Int16,
            'iceTime': pl.Int16,
            'goals': pl.Int16,
            'xGoals': pl.Float32,
            'lowDangerShots': pl.Int16,
            'mediumDangerShots': pl.Int16,
            'highDangerShots': pl.Int16,
            'lowDangerxGoals': pl.Float32,
            'mediumDangerxGoals': pl.Float32,
            'highDangerxGoals': pl.Float32,
            'lowDangerGoals': pl.Int16,
            'mediumDangerGoals': pl.Int16,
            'highDangerGoals': pl.Int16,
        }
    )

    final_df = final_df.sort(by=['team', 'playerID'])

    return final_df



import polars as pl

from pyhockey.team_games import team_games


def test_standard_team_games():
    """
    Test that a standard request from team_games gives a DF of the expected shape.
    """
    result: pl.DataFrame = team_games(team='TOR', start_date='2024-10-01',
                                      end_date='2025-01-23')

    assert result.shape == (49, 15)



"""
Module for handling calls to the 'goalies' table, which holds season summaries for each goalie,
divided by game state.
"""

import polars as pl

from pyhockey.util.query_table import query_table
from pyhockey.util.data_disclaimer import print_data_disclaimer


# Define custom type for inputs into our queries
type QueryValue = str | int | float | list[str] | list[int] | list[float]


def goalie_summary(season: int | list[int],
                   team: str | list[str] = 'ALL',
                   min_games_played: int = 0,
                   situation: str = 'all',
                   combine_seasons: bool = False,
                 quiet: bool = False) -> pl.DataFrame:
    """
    Primary function for retrieving goalie-level season summaries. Given a season or list of
    seasons, return goalie data summaries for each of those seasons. 

    Can provide further filters via a team or list of teams, a minimum icetime cutoff, or
    a specific situation/game state.

    :param int | list[int] season: The (list of) season(s) for which to return data
    :param str | list[str] team: The (list of) team(s) for which to return data, defaults to 'ALL'
    :param int min_icetime: A minimum icetime (in minutes) cut-off to apply, defaults to 0
    :param str situation: One of 'all', '5on5', '4on5', '5on4', or 'other', defaults to 'all'
    :param bool combine_seasons: If True, and given multiple seasons, combine the results of each
                                 season into a single entry for each player, defaults to False
    :param bool quiet: If set to True, don't print the data disclaimer, defaults to False

    :return pl.DataFrame: The resulting data in a polars DataFrame
    """

    column_mapping: dict[str, QueryValue] = {
        'season': season,
        'team': team,
        'situation': situation
    }

    qualifiers: dict[str, str] = {
        'gamesPlayed': f'>={min_games_played}'
    }

    results: pl.DataFrame = query_table(table='goalies', column_mapping=column_mapping,
                                        qualifiers=qualifiers, combine_seasons=combine_seasons,
                                        order_by=['team', 'season'])

    if not quiet:
        print_data_disclaimer(source='MoneyPuck')

    return results



import pytest

from pyhockey.util.input_validation import check_input_type, check_input_values, \
validate_date_range


def test_check_valid_inputs_singleton_success():
    """
    Checks to see that valid singleton inputs do not raise any errors.
    """
    test_inputs: dict[str] = {
        'season': 2024,
        'team': 'TOR',
        'situation': '5on5'
    }

    result: bool = check_input_values(test_inputs)

    assert result


def test_check_valid_inputs_list_success():
    """
    Checks to see that valid list inputs do not raise any errors.
    """
    test_inputs: dict[str] = {
        'season': [2023, 2024, 2025],
        'team': ['TOR', 'MTL', 'OTT'],
        'situation': '5on5'
    }

    result: bool = check_input_values(test_inputs)

    assert result


def test_check_valid_inputs_singleton_failure():
    """
    Checks to see that invalid singleton inputs raise errors as expected
    """
    test_inputs: dict[str] = {
        'season': 1999,
        'team': ['TOR', 'MTL', 'OTT'],
        'situation': '5on5'
    }

    with pytest.raises(ValueError):
        check_input_values(test_inputs)


def test_check_valid_inputs_list_failure():
    """
    Checks to see that invalid list inputs raise errors as expected
    """
    test_inputs: dict[str] = {
        'season': 2024,
        'team': ['TOR', 'FOO', 'OTT', 'BAR'],
        'situation': '5on5'
    }

    with pytest.raises(ValueError):
        check_input_values(test_inputs)


def test_check_input_type_singleton_success():
    """
    Test that check_input_type() works as expected with single strings and ints.
    """
    test_inputs: dict[str] = {
        'season': 2024,
        'team': 'TOR',
        'situation': '5on5'
    }

    result: bool = check_input_type(column_mapping=test_inputs)

    assert result


def test_check_input_type_singleton_failure():
    """
    Test that check_input_type() raises a ValueError when mismatched singletons are provided.
    """
    test_inputs: dict[str] = {
        'season': '2024',
        'team': 'TOR',
        'situation': '5on5'
    }

    with pytest.raises(ValueError):
        check_input_type(test_inputs)


def test_check_input_type_list_success():
    """
    Test that check_input_type() works as expected with lists of strings and ints.
    """
    test_inputs: dict[str] = {
        'season': [2023, 2024, 2025],
        'team': ['TOR', 'MTL', 'OTT'],
        'situation': '5on5'
    }

    result: bool = check_input_type(test_inputs)

    assert result


def test_check_input_type_list_failure():
    """
    Test that check_input_type() fails as expected when given a list with at least one incorrect
    type.
    """
    test_inputs: dict[str] = {
        'team': ['TOR', 'MTL', 'OTT'],
        'situation': '5on5',
        'season': [2023, 2024, '2025']
    }

    with pytest.raises(ValueError):
        check_input_type(test_inputs)


def test_date_range_validation_start_and_end():
    """
    Test that validate_date_range() works as expected when both a start_date and end_date are
    provided.
    """
    column_mapping = {
        'season': 2025,
        'situation': 'all'
    }

    qualifiers = {
        'start_date': '2025-10-10',
        'end_date': '2025-10-30'
    }

    expected = {
        'situation': 'all'
    }

    result = validate_date_range(column_mapping=column_mapping, qualifiers=qualifiers)

    assert result == expected
   

def test_date_range_validation_start_and_season():
    """
    Test that validate_date_range() works as expected when just a start_date is provided in
    addition to a season.
    """
    column_mapping = {
        'season': 2025,
        'situation': 'all'
    }

    qualifiers = {
        'start_date': '2025-10-10',
    }

    result = validate_date_range(column_mapping=column_mapping, qualifiers=qualifiers)

    assert result == column_mapping


def test_date_range_validation_raises_error_on_bad_date_format():
    """
    Test that validate_date_range() raises a ValueError when a date range in an incorrect
    format is provided.
    """
    column_mapping = {
        'season': 2025,
        'situation': 'all'
    }

    qualifiers = {
        'start_date': '10-10-2025',
        'end_date': '2025-10-30'
    }

    with pytest.raises(ValueError):
        validate_date_range(column_mapping=column_mapping, qualifiers=qualifiers)


def test_date_range_validation_raises_error_on_end_before_start():
    """
    Test that validate_date_range() raises a ValueError when an end_date that comes before the
    start_date is provided
    """
    column_mapping = {
        'season': 2025,
        'situation': 'all'
    }

    qualifiers = {
        'start_date': '2025-10-30',
        'end_date': '2024-10-30'
    }

    with pytest.raises(ValueError):
        validate_date_range(column_mapping=column_mapping, qualifiers=qualifiers)



MOTHERDUCK_READ_TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6InNvaHJhdWJwYXp1a2lAZ21haWwuY29tIiwic2Vzc2lvbiI6InNvaHJhdWJwYXp1a2kuZ21haWwuY29tIiwicGF0IjoiSklBM2UwLXctUkxUWXRoYXR0b0s5MnUxWXAxX0I3bGxzWjNIUkR0ZzRmVSIsInVzZXJJZCI6IjYzMzFhOGM5LWQ4YTctNDQzOS1iMjI4LTYzMTEzMTM2MWU5NyIsImlzcyI6Im1kX3BhdCIsInJlYWRPbmx5Ijp0cnVlLCJ0b2tlblR5cGUiOiJyZWFkX3NjYWxpbmciLCJpYXQiOjE3NTk3ODA5Nzh9.tYlhhvZTGJLODSVJAIzXPZx1bz5xoROqhvOTfH_qAqU'



import polars as pl

from pyhockey.team_summary import team_summary


def test_standard_team_summary():
    """
    Test that a standard request from team_summary gives a DF of the proper shape.    
    """
    result: pl.DataFrame = team_summary(season=2023)

    assert result.shape == (32, 13)


def test_combined_team_summary():
    """
    Test that a request using combined_seasons = True gives a DF of the proper shape.
    """
    result: pl.DataFrame = team_summary(season=[2023, 2024], combine_seasons=True)

    assert result.shape == (33, 14)
